"""
 Python functions for solving and plotting the solution to
 an Assortative Matching with Large Firms (2016) model (with capital)

@author : Cristina Lafuente
@date : 31/05/2016
"""

# Generic modules
from __future__ import division
import numpy as np
import sympy as sym
from numpy.polynomial.polynomial import polyval
from scipy.optimize import leastsq, fsolve,least_squares
from scipy.interpolate import interp1d
import operator
from scipy.special import erf
# Plotting modules
import matplotlib.pyplot as plt
import seaborn as sn
sn.set_style('whitegrid')
cp = sn.xkcd_rgb["pinkish"]
cb = "#3498db"
cr = "#1fa774"
# Specific modules
import inputs
import models
import shooting

_modules = [{'ImmutableMatrix': np.array, 'erf': erf, 'sqrt': np.sqrt}, 'numpy']

x, y = sym.var('x, y')
R, l, r, rho, gamma, eta, L, A, kapa, p = sym.var('R, l, r, rho, gamma, eta, L, A, kapa, p')
x0,x1= [3.16e-6,1e2]

# PART I : SOLVING FUNCTIONS

def get_k(country, ftype=0, fchoice=None,plot=False, save_in=None, x_range=[0.999,1.001]):
	"""
	Calcualtes three approximated functions of k(x,y,l,r) and returns the one with the lowest MSE.

	WARNING: get_k() (and the solver) use an inverted production function, such that:
		* output xs are model ys
		* output ys and model xs and
		* output thetas are model 1/l
		
	So the output k is actually defined as k(y,x,1/l,1/r)	

	Inputs
	------
	country	: (str) "R" or "P" for rich (US baseline) or poor (Aggregate Factors) country parameters
	ftype 	: (int) Type of function. Possible values:
					* 0 : x as an exponent of (l/r)
					* 1 : x multiplying k
					* 2 : x as a scaling factor (multiplying A)
					* 3 : x multiplying k (but not raised to rho)
	fchoice	: (int) Type of approximation function to k. If None, choses best fit measured by MSE.
					Values:
					* 0 for cuadratic tensors
					* 1 for cubic tensors
					* 2 for polynomial on x(r/l)^y (WARNING: not well behaved)

	plot 	: (bool) True for producing a plot of the fit of the three kinds
	save 	; (bool) True for saving the paramers. File names will be:
					- "pams_new.csv" -> Parameters of the polynomial on (x(r/l)^y)
					- "pams3.csv" -> Parameters of the cubic tensors
					- "pams.csv" -> Parameters of the quadratic tensors
	Output
	------
	k_star	: (sym) Symbolical expression for k(x,y,l,r)

	"""

	if country=="R":
		y3d = np.linspace(x_range[0],x_range[1],11)
		x3d = np.logspace(-6,2.0,101)
		l3d = np.logspace(1,np.log10(1/5000),101)

		ygrid, xgrid, lgrid = np.meshgrid(y3d,x3d,l3d)

		if ftype==0:
			p = -(0.89**2*1.0*1.0*0.3159)/(2*0.13099)
			q = -(0.89*(1-0.89)*1.0*1.0*0.3159*(xgrid*(1/lgrid)**ygrid)**(1/4))/(2*0.13099)
		# With A
		elif ftype==1:
			p = -(0.89**2*1.0*1.0*ygrid*0.3159)/(2*0.13099)
			q = -(0.89*(1-0.89)*1.0*ygrid*1.0*0.3159*(xgrid*(1/lgrid))**(1/4))/(2*0.13099)
		# (xk)^(1/4)
		elif ftype==2:
			p = -(0.89**2*1.0*1.0*0.3159*ygrid**(1/2))/(2*0.13099)
			q = -(0.89*(1-0.89)*1.0*1.0*0.3159*(xgrid*(1/lgrid)*ygrid)**(1/4))/(2*0.13099)
		# x(k)^(1/4)
		elif ftype==3:
			p = -(0.89**2*1.0*1.0*0.3159*ygrid**2)/(2*0.13099)
			q = -(0.89*(1-0.89)*1.0*1.0*ygrid*0.3159*(xgrid*(1/lgrid))**(1/4))/(2*0.13099)
		elif ftype==4:
			pass
		else:
			raise ValueError("Please specify a valid function type (0-4")

		if ftype!=4:		
			pq = 4*p**3+27*q**2

			filter1 = pq>0 #root1
			filter3 = 1-filter1

			q1 = abs(filter1*q)
			q3 = filter3*q

			bit1 = (2*np.sqrt((-p/3)) * np.cosh( (1/3)*np.arccosh( ((-3*abs(q1))/(2*p)) * np.sqrt((-3/p)) ) ))**4
			bit1 = filter1*bit1
			bit1 = np.nan_to_num(bit1)
			bit3 = (2*np.sqrt((-p/3))*np.cos( (1/3)*np.arccos( ((3*q3)/(2*p)) * np.sqrt((-3/p)) ) ))**4
			bit3 = filter3*bit3

			Kgrid = bit1+bit3
		else:
			Kgrid = solve_for_kX(country) # see below
	
	elif country=="P":
		y3d = np.linspace(x_range[0],x_range[1],11)
		x3d = np.logspace(-6,2.0,101)
		l3d2 = np.logspace(1.2,-3,101)

		ygrid, xgrid, lgrid = np.meshgrid(y3d,x3d,l3d2)

		if ftype==0:
			p2 = -(0.89**2*1.0*0.3987*0.5209)/(2*0.3958)
			q2 = -(0.89*(1-0.89)*1.0*0.3987*0.5209*(xgrid*(1/lgrid)**ygrid)**(1/4))/(2*0.3958)
		# With A
		elif ftype==1:
			p2 = -(0.89**2*1.0*ygrid*0.3987*0.5209)/(2*0.3958)
			q2 = -(0.89*(1-0.89)*1.0*ygrid*0.3987*0.5209*(xgrid*(1/lgrid))**(1/4))/(2*0.3958)
		# With (kx)^1/4
		elif ftype==2:
			p2 = -(0.89**2*1.0*0.3987*0.5209*ygrid**(1/2))/(2*0.3958)
			q2 = -(0.89*(1-0.89)*1.0*0.3987*0.5209*(xgrid*(1/lgrid)*ygrid)**(1/4))/(2*0.3958)
		# With x(k)^1/4
		elif ftype==3:
			p2 = -(0.89**2*1.0*0.3987*0.5209*ygrid**2)/(2*0.3958)
			q2 = -(0.89*(1-0.89)*1.0*0.3987*ygrid*0.5209*(xgrid*(1/lgrid))**(1/4))/(2*0.3958)
		elif ftype==4:
			pass
		else:
			raise ValueError("Please specify a valid function type (0-4")

		if ftype!=4:		
			pq2 = 4*p2**3+27*q2**2

			filter12 = pq2>0 #root1
			filter32 = 1-filter12

			q12 = abs(filter12*q2)
			q32 = filter32*q2

			bit12 = (2*np.sqrt((-p2/3)) * np.cosh( (1/3)*np.arccosh( ((-3*abs(q12))/(2*p2)) * np.sqrt((-3/p2)) ) ))**4
			bit12 = filter12*bit12
			bit12 = np.nan_to_num(bit12)
			bit32 = (2*np.sqrt((-p2/3))*np.cos( (1/3)*np.arccos( ((3*q32)/(2*p2)) * np.sqrt((-3/p2)) ) ))**4
			bit32 = filter32*bit32

			Kgrid = bit12+bit32
		
		else:
			Kgrid = solve_for_kX(country) # see below
	else:
		raise ValueError("Please specify R for rich or P for poor country parameters")
	
	def res_ten3(params, out, arg1, arg2, arg3):   
		kr, z, a, b = out.flatten(), arg1.flatten(), arg2.flatten(), arg3.flatten()
		a3b3z3, z3, a3, b3, ab3, az3, bz3, zb3, ba3, za3, z1, z2, abz, a2z, b2z, bz2, az2 ,a2bz2,ab2z2,a2b2z, lz, lza, lzb, a1,a2,b1,b2,ab,a2b,ab2,a2b2,la,lb,lab = params
		return a3b3z3*a**3*(1/b)**3*z**3 + z3*z**3 + a3*a**3 + b3*(1/b)**3 + ab3*a*(1/b)**3 + az3*a*z**3 + bz3*(1/b)*z**3 + zb3*z*(1/b)**3 + ba3*(1/b)*a**3 + za3*z*a**3 + z1*z + z2*z**2 + abz*a*(1/b)*z + a2z*a**2*z + b2z*(1/b)**2*z + bz2*(1/b)*z**2 + az2*a*z**2 + a2bz2*a**2*(1/b)*z**2 + ab2z2*a*(1/b)**2*z**2 + a2b2z*a**2*(1/b)**2*z + lz*np.log(z) + lza*np.log(z)*np.log(a) + lzb*np.log(z)*np.log(1/b) + a1*a + a2*a**2 + b1*(1/b) + b2*(1/b)**2 + ab*a*(1/b) + a2b*a**2*(1/b) + ab2*a*(1/b)**2 + a2b2*a**2*(1/b)**2 + la*np.log(a) + lb*np.log(1/b) + lab*np.log(a)*np.log(1/b) - kr
	# Log tensors of the second kind
	def res_ten3o(params, out, arg1, arg2, arg3):
		kr, z, a, b = out.flatten(), arg1.flatten(), arg2.flatten(), arg3.flatten()
		cons,z1, z2, abz, a2z, b2z, bz2, az2 ,a2bz2,ab2z2,a2b2z, lz, lza, lzb, a1,a2,b1,b2,ab,a2b,ab2,a2b2,la,lb,lab = params
		return cons + z1*z + z2*z**2 + abz*a*(1/b)*z + a2z*a**2*z + b2z*(1/b)**2*z + bz2*(1/b)*z**2 + az2*a*z**2 + a2bz2*a**2*(1/b)*z**2 + ab2z2*a*(1/b)**2*z**2 + a2b2z*a**2*(1/b)**2*z + lz*np.log(z) + lza*np.log(z)*np.log(a) + lzb*np.log(z)*np.log(1/b) + a1*a + a2*a**2 + b1*(1/b) + b2*(1/b)**2 + ab*a*(1/b) + a2b*a**2*(1/b) + ab2*a*(1/b)**2 + a2b2*a**2*(1/b)**2 + la*np.log(a) + lb*np.log(1/b) + lab*np.log(a)*np.log(1/b) - kr
	def pol_shorcut(params,exact_k,xs,ys,ls):
		if ftype==0:
			rosebud = (xs*(1.0/ls)**ys)
			return polyval(rosebud,params[:-1]) + params[-1]*np.log(rosebud) - exact_k
		else:
			rosebud = (xs*(1.0/ls))*ys
			return polyval(rosebud,params[:22]) + params[22]*np.log(rosebud) + polyval(ys,params[23:-1]) + params[-1]*np.log(ys) - exact_k

	params_ten3 = leastsq(res_ten3,  np.zeros(34), args=(Kgrid,xgrid,ygrid,lgrid))[0]
	params_s = leastsq(pol_shorcut,  np.zeros(34), args=(Kgrid.flatten(),xgrid.flatten(),ygrid.flatten(),lgrid.flatten()))[0]
	params_ten3o = leastsq(res_ten3o,  np.zeros(25), args=(Kgrid,xgrid,ygrid,lgrid))[0]

	mse_teno = np.sum((Kgrid.flatten() - res_ten3o(params_ten3o,np.array([0.,]),xgrid,ygrid,lgrid))**2)/len(Kgrid.flatten())
	mse_ten3 = np.sum((Kgrid.flatten()- res_ten3(params_ten3,np.array([0.,]),xgrid,ygrid,lgrid))**2)/len(Kgrid.flatten())
	mse_s = np.sum((Kgrid.flatten() - pol_shorcut(params_s,np.array([0.,]),xgrid.flatten(),ygrid.flatten(),lgrid.flatten()))**2)/len(Kgrid.flatten())

	if fchoice==None:
		best_fit = np.argmin(np.array((mse_teno,mse_ten3,mse_s)))
	else:
		best_fit=fchoice

	if best_fit==0:
		cons,c1,c2,abc,a2c,b2c,bc2,ac2,a2bc2,ab2c2,a2b2c,lc,lca,lcb,a1,a2,b1,b2,ab,a2b,ab2,a2b2,la,lb,lab = params_ten3o
		k_star = cons + c1*x + c2*x**2 + abc*x*y*(r/l) + a2c*y**2*x + b2c*(r/l)**2*x + bc2*(r/l)*x**2 + ac2*y*x**2 + a2bc2*y**2*(r/l)*x**2 + ab2c2*y*(r/l)**2*x**2 + a2b2c*y**2*(r/l)**2*x + lc*sym.log(x) + lca*sym.log(x)*sym.log(y) + lcb*sym.log(x)*sym.log(r/l) + a1*y + a2*y**2 + b1*(r/l) + b2*(r/l)**2 + ab*y*(r/l) + a2b*y**2*(r/l) + ab2*y*(r/l)**2 + a2b2*y**2*(r/l)**2 + la*sym.log(y) + lb*sym.log(r/l) + lab*sym.log(r/l)*sym.log(y)
	elif best_fit==1:
		a3b3c3,c3,a3,b3,ab3,ac3,bc3,cb3,ba3,ca3,c1,c2,abc,a2c,b2c,bc2,ac2,a2bc2,ab2c2,a2b2c,lc,lca,lcb,a1,a2,b1,b2,ab,a2b,ab2,a2b2,la,lb,lab = params_ten3
		k_star = a3b3c3*y**3*(r/l)**3*x**3 + c3*x**3 + a3*y**3 + b3*(r/l)**3 + ab3*y*(r/l)**3 + ac3*y*x**3 + bc3*(r/l)*x**3 + cb3*x*(r/l)**3 + ba3*(r/l)*y**3 + ca3*x*y**3 + c1*x + c2*x**2 + abc*x*y*(r/l) + a2c*y**2*x + b2c*(r/l)**2*x + bc2*(r/l)*x**2 + ac2*y*x**2 + a2bc2*y**2*(r/l)*x**2 + ab2c2*y*(r/l)**2*x**2 + a2b2c*y**2*(r/l)**2*x + lc*sym.log(x) + lca*sym.log(x)*sym.log(y) + lcb*sym.log(x)*sym.log(r/l) + a1*y + a2*y**2 + b1*(r/l) + b2*(r/l)**2 + ab*y*(r/l) + a2b*y**2*(r/l) + ab2*y*(r/l)**2 + a2b2*y**2*(r/l)**2 + la*sym.log(y) + lb*sym.log(r/l) + lab*sym.log(r/l)*sym.log(y)
	else:
		if ftype == 0 :
			p33, p32, p31, p30, p29, p28, p27, p26, p25, p24, p23, p22, p21, p20, p19, p18, p17, p16, p15, p14, p13, p12, p11, p10, p9, p8, p7, p6, p5, p4, p3, p2, p1, p34 = params_s
			Z = (x*(r/l)**y)
			k_star = p1 + p2*Z + p3*Z**2 + p4*Z**3 + p5*Z**4 + p6*Z**5 + p7*Z**6 + p8*Z**7 + p9*Z**8 + p10*Z**9 + p11*Z**10 + p12*Z**11 + p13*Z**12 + p14*Z**13 + p15*Z**14 + p16*Z**15 + p17*Z**16 + p18*Z**17 + p19*Z**18 + p20*Z**19 + p21*Z**20 + p22*Z**21 + p23*Z**22 + p24*Z**23 + p25*Z**24 + p26*Z**25 + p27*Z**26 + p28*Z**27 + p29*Z**28 + p30*Z**29 + p31*Z**30 + p32*Z**31 + p33*Z**32 + p34*sym.log(Z)
		else:
			p21, p20, p19, p18, p17, p16, p15, p14, p13, p12, p11, p10, p9, p8, p7, p6, p5, p4, p3, p2, p1, p22, p33, p32, p31, p30, p29, p28, p27, p26, p25, p24, p23, p34 = params_s
			Z = (x*(r/l)*y)
			k_star = p1 + p2*Z + p3*Z**2 + p4*Z**3 + p5*Z**4 + p6*Z**5 + p7*Z**6 + p8*Z**7 + p9*Z**8 + p10*Z**9 + p11*Z**10 + p12*Z**11 + p13*Z**12 + p14*Z**13 + p15*Z**14 + p16*Z**15 + p17*Z**16 + p18*Z**17 + p19*Z**18 + p20*Z**19 + p21*Z**20 + p22*sym.log(Z) + p23*y + p24*y**2 + p25*y**3 + p26*y**4 + p27*y**5 + p28*y**6 + p29*y**7 + p30*y**8 + p31*y**9 + p32*y**10 + p33*y**11 + p34*sym.log(y)
	
	# saving module
	if save_in!=None:
		if best_fit==0:
			np.savetxt(save_in+"0", params_ten3o, delimiter=",")
		elif best_fit==1:
			np.savetxt(save_in+"1", params_ten3, delimiter=",")
		else:
			np.savetxt(save_in+"c", params_s, delimiter=",")		
		
	#plotting module
	if plot:
		gridd = Kgrid, xgrid, ygrid, lgrid
		paramss = params_ten3o,params_ten3,params_s
		fs = res_ten3o, res_ten3, pol_shorcut
		plot_k_fit(gridd,fs,paramss,best_fit)
	
	return k_star



def solve(country, k_sol, ftype=0, logs=True, spread='even',dispams=None,x_range=[0.99999,1.00001],
	assort="positive", verbose=False,guess_overide=1.0,scaling_x=1.0,normconst=0.0):
	"""

	Solves a 2-sided heterogeneity matching model usign a shooting algorithm, given capital choice k.

	WARNING: the solver (and get_k()) use an inverted production function, such that:
		* output xs are model ys
		* output ys and model xs and
		* output thetas are model 1/l 

	Inputs
	------
	country (str)			: "R" or "P" for Rich or Poor country parameters
	k_sol (sympy expression):

	Optional
	........

	ftype (int)				: Specification of production function (default=0) (see "Introduction" in the notebook)
	logs (bool)				: True (default) for log steps.
	spread (str)			: 'even' (default) for uniform distribution of x
							  'lognormal' for lognormal distribution of x
							  'normal' for normal distribution of x
							  'bimod' for bimodal distribution of x (2 join normals with same variance and different means)
	disparams (list or tup) : in case that the spread is not even, refers to values of the mean and std of the distribtuions;
							  for the bimodal, it must have length=3 and contains mean1, mean2 and the common variance;
							  all distribtuions have defauls parameters if the user does not specify hers.
	x_range (list or tup)	: range for x
	assort (str)			: if "positive" (default) assumes (and forces) PAM, "negative" does NAM
	verbose	(bool)			: if True prints progress of the solver; False is default.
	guess_overide (float)	: if the user wants to start with a particular inital guess for maximum land size. Default is 1.
	scaling_x (float)		: scalar that multiplies the cdf of x, used to make sure it integrates to 1.
	normconst (float)		: scalar to noramlize the cdf of x, used to ensure cdf_x(max(x))=1

	Outputs
	-------
	A solution dictionary with he keys:
	"solved" (bool)					: True is the solver was successful
	"xs","ys","thetas",
	"ws","pis","ks" (numpy.arrays)	: arrays with the points along the solution path 
	"solution": (pandas.dataframe)	: all of the above in a dataframe for convinience
	"pdf_dis" (numpy.array)			: the discretized values of the pdf(y)
	"F" (sympy expression)			: Production function chosen above. Convient for calculating derivatives (below)
	"PAM" (str)						: Assortativity ('positive' or 'negative')

	"""
	x0,x1= [3.16e-6,1e2]
	y0,y1=x_range

	if country=="R":
		F_params = {'R':0.13099, 'rho':0.25, 'gamma':0.5, 'eta':0.890204456766942, 'A': 1.0, 'kapa':1.0, 'p':0.3159}
		x_scale = 169.249 * scaling_x
		k_star = k_sol
		guess = guess_overide
	else:
		F_params = {'R':0.3958, 'rho':0.25, 'gamma':0.5, 'eta':0.890204456766942, 'A': 0.3987, 'kapa':1.0, 'p':0.5209}
		x_scale = 19.595 * scaling_x
		k_star = k_sol
		guess = guess_overide*3


	# Firms
	if spread=='even':
		y, a, b = sym.var('y, a, b')
		skill_cdf = ((y - a) / (b - a)) * x_scale
		skill_params = {'a': y0, 'b': y1}
		skill_bounds = [y0, y1]
	elif spread=='lognorm':
		y, loc1, mu1, sigma1 = sym.var('y, loc1, mu1, sigma1')
		skill_cdf = (0.5 + 0.5 * sym.erf((sym.log(y - loc1) - mu1) / sym.sqrt(2 * sigma1**2)))* x_scale - normconst
		if dispams==None:
			skill_params = {'loc1': 0.0, 'mu1': -0.03, 'sigma1': 0.055}
		else:
			skill_params = {'loc1': 0.0, 'mu1': dispams[0], 'sigma1': dispams[1]}
		skill_bounds = [y0, y1]
	elif spread=='norm':
		sig1, y, mu1 = sym.var("sig1, y, mu1")
		skill_cdf = 0.5 * (1 + sym.erf((y-mu1)/ sym.sqrt(2 * sig1**2))) * x_scale - normconst
		if dispams==None:
			skill_params = {'mu1': 1.1, 'sig1': 0.05 }
		else:
			skill_params = {'mu1': dispams[0], 'sig1': dispams[1]}
		skill_bounds = [y0, y1]
	elif spread=='bimod':
		sig1, sig2, y, mu1, mu2= sym.var("sig1, sig2, y, mu1, mu2")
		c1 = 0.5 * (1 + sym.erf((y-mu1)/ sym.sqrt(2 * sig1**2)))
		c2 = 0.5 * (1 + sym.erf((y-mu2)/ sym.sqrt(2 * sig2**2)))
		skill_cdf  = (c1*0.5+c2*0.5) * x_scale - normconst
		if dispams==None:
			skill_params = {'mu1': 0.9, 'mu2': 1.1, 'sig1': 0.05,'sig2': 0.05}
		else:
			skill_params = {'mu1': dispams[0], 'mu2': dispams[1], 'sig1': dispams[2],'sig2': dispams[2]}
		skill_bounds = [y0, y1]
		
	else:
		raise ValueError, 'Unvalid spread. Choose even or lognorm.'

	firms = inputs.Input(var=y,cdf=skill_cdf,params=skill_params,bounds=skill_bounds)

	# Workers
	x, loc2, mu2, sigma2 = sym.var('x, loc2, mu2, sigma2')
	productivity_cdf = (0.5 + 0.5 * sym.erf((sym.log(x - loc2) - mu2) / sym.sqrt(2 * sigma2**2)))
	productivity_params = {'loc2': 0.0, 'mu2': -1.8316, 'sigma2': 4.6553}
	productivity_bounds = (x0, x1)

	workers = inputs.Input(var=x,cdf=productivity_cdf,params=productivity_params,bounds=productivity_bounds)

	# Production Function
	if ftype==0:
		F = (l)*p*A*kapa*(eta*(k_star)**rho + (1- eta)*(((r/l)**y)*x)**rho)**(gamma/rho) - (l)*R*k_star
	elif ftype==1: # xA
		F = y*(l)*p*A*kapa*(eta*(k_star)**rho + (1- eta)*((r/l)*x)**rho)**(gamma/rho) - (l)*R*k_star
	elif ftype==2: # xk
		F = (l)*p*A*kapa*(eta*(k_star*y)**rho + (1- eta)*((r/l)*x)**rho)**(gamma/rho) - (l)*R*k_star
	elif ftype==3: # x(k^1/4)
		F = (l)*p*A*kapa*(eta*y*(k_star)**rho + (1- eta)*((r/l)*x)**rho)**(gamma/rho) - (l)*R*k_star
	elif ftype==4: # k^x
		F = (l)*p*A*kapa*(eta*(k_star**y)**rho + (1- eta)*((r/l)*x)**rho)**(gamma/rho) - (l)*R*k_star

	#Setting up the solver
	model = models.Model(assortativity=assort,workers=workers,firms=firms,production=F,params=F_params)
	solver = shooting.ShootingSolver(model=model)

	# log steps
	if logs:
		nk = None
		kts = np.logspace(np.log10(x0),np.log10(x1),num=6000)
	# Uniform steps
	elif not logs:
		nk =6000
		kts = None

	# Solve
	try:
		solver.solve(guess, tol=1e-9, number_knots=nk, knots=kts,integrator="lsoda", message=verbose)
	except AssertionError, e1:
		try:
			solver.solve(guess*10, tol=1e-9, number_knots=nk, knots=kts,integrator='lsoda', message=verbose)
		except ValueError:
			try:
				solver.solve(guess/10, tol=1e-9, number_knots=nk, knots=kts,integrator='lsoda', message=verbose)
			except ValueError:
				pass
		except AssertionError:
			try:
				solver.solve(guess/10, tol=1e-9, number_knots=nk, knots=kts,integrator='lsoda', message=verbose)
			except ValueError:
				pass
			except AssertionError:
				try:
					solver.solve(guess*10, tol=1e-9, number_knots=nk, knots=kts,integrator='lsoda', message=verbose)
				except ValueError:
					pass
				except AssertionError:
					pass

	except ValueError:
		try:
			solver.solve(guess*10, tol=1e-9, number_knots=nk, knots=kts,integrator='lsoda', message=verbose)
		except AssertionError:
			try:
				solver.solve(guess/10, tol=1e-9, number_knots=nk, knots=kts,integrator='lsoda', message=verbose)
			except ValueError:
				pass
		except ValueError:
			pass
 
	# Store
	try:
		sol1 = solver.solution
	except AttributeError:
		return {'solved': False}
	xs = sol1.index.values
	ys = sol1['$\\mu(x)$'].values
	thetas = sol1['$\\theta(x)$'].values
	ws = sol1['$w(x)$'].values
	pis = sol1['$\\pi(x)$'].values

	# For capital lambdify k_star
	eval_k = sym.lambdify((x,y,l,r), k_sol, modules='numpy')
	ks = eval_k(xs,ys,thetas,1.0)

	def discrete_pdf(x0,x1):
		p_cdf = productivity_cdf.subs({'loc2': 0.0, 'mu2': -1.8316, 'sigma2': 4.6553})
		y_cdf = sym.lambdify(x,p_cdf,modules = [{'ImmutableMatrix': np.array, 'erf': erf}, 'numpy'])
		return y_cdf(x1)-y_cdf(x0)
	
	xdist = (xs[1:]-xs[:-1])/2
	xdist = np.hstack((xdist[0],xdist))
	xsd = discrete_pdf(xs-xdist,xs+xdist)
	pdf_dis = xsd/np.sum(xsd)

	return {"solved":len(xs)==6000,"xs":xs , "ys":ys , "thetas":thetas , "ws": ws, "pis": pis, "ks": ks, "solution": sol1, "pdf_dis":pdf_dis, "F":F, "PAM": assort}

# PART II: Auxiliary functions # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # 

def solve_for_kX(country):
	"""
	Auxiliary function to solve for the case when k^x.
	"""
	# Setting parameters
	kappa = 1.0
	eta = 0.89
	r = 1
	if country=="R":
		A = 1.0
		R = 0.13099
		pa = 0.3159
	else:
		A = 0.3987
		R = 0.3958
		pa = 0.5209
	# FOC
	def f_k(k,x,y,l):
		return x*A*pa*kappa*eta**2*k**(2*(x/4)) + x*A*kappa*eta*(1-eta)*(y*(l/r))**(1/4)*k**(x/4)-2*R*k
	# Create grid, solving one by one
	k3d = np.zeros((len(y3d),len(x3d),len(l3d)))
	xi = 0
	yi = 0
	li = 0
	for y in y3d:
	    for x in x3d:
	        for l in l3d:
	            k3d[yi,xi,li] = fsolve(f_k, 10.0*(1/l), args=(y,x,1/l))[0]
	            li +=1
	        li = 0
	        xi +=1
	    xi = 0
	    yi+=1
	return k3d

def derivatives(solution, country):
	"""
	Calcualtes the value of the cross-derivates of the production function at the solution points given.

	Convient for checking that the PAM condition holds: Fxy Flr > Fly Fxr (plot_PAM() below)

	Inputs
	------
	solution (dic)	: a dictionary from solve() above.
	country (str)	: "R" or "P" for the US or poor country baseline

	Outputs
	-------
	A dictionary with the derivatives:
		- Fxy
		- Flr
		- Fyl
		- Fxr
	"""
	# Unpacking
	F = solution["F"]
	xs,ys,thetas = solution["xs"],solution["ys"],solution["thetas"]
	R, l, r, rho, gamma, eta, L, A, kapa, p = sym.var('R, l, r, rho, gamma, eta, L, A, kapa, p')
	# Derivatives
	Fxy = sym.lambdify((x,y,l,r,A,kapa,eta,gamma,rho,R,p),F.diff(x,y), modules='numpy')
	Flr = sym.lambdify((x,y,l,r,A,kapa,eta,gamma,rho,R,p),F.diff(l,r), modules='numpy')
	Fyl = sym.lambdify((x,y,l,r,A,kapa,eta,gamma,rho,R,p),F.diff(y,l), modules='numpy')
	Fxr = sym.lambdify((x,y,l,r,A,kapa,eta,gamma,rho,R,p),F.diff(x,r), modules='numpy')

	if country=="P":
		R,A,p = 0.3958,0.3987,0.5209
	elif country=="R":
		R,A,p = 0.13099,1.0,0.3159
	dxy = Fxy(xs,ys,thetas,1.0,A,1.0,0.89,0.5,0.25,R,p)
	dlr = Flr(xs,ys,thetas,1.0,A,1.0,0.89,0.5,0.25,R,p)
	dyl = Fyl(xs,ys,thetas,1.0,A,1.0,0.89,0.5,0.25,R,p)
	dxr = Fxr(xs,ys,thetas,1.0,A,1.0,0.89,0.5,0.25,R,p)
	
	return {"Fxy": dxy, "Flr": dlr, "Fyl": dyl, "Fxr":dxr}

def get_kl_ratio(land,capital,bins,pdf_dis):
	"""
	Calculates the capital to land ratio distribution, given a set of bins, normalized by the highest bin.

	Convinient for replicating the histogram in A&R.

	Inputs
	-----
	land (np.array)		: array of equilibrium land choices (form solve())
	capital (np.array)	: array of equilibrium capital choices (form solve())
	bins (list or array): increasing array with bin edges
	pdf_dis (np.array)	: array with the weights of each point (form solve()), same length as land and capital

	Output
	------
	np.array with the values of the average k/l ratio for each bin, divided by that of the last bin.
	"""
	total_land = np.sum(land*pdf_dis)
	total_land_m = np.sum(farm_grid*y_prob)

	land_bin = np.zeros(13)
	k_bin = np.zeros(13)
	land_bin_m = np.zeros(13)
	k_bin_m = np.zeros(13)
	kl_bar = np.zeros(13)
	kl_bar_model = np.zeros(13)

	linx = 0
	b = 0
	for i in range(len(ks)):    
		if land[i]>=bins[b]:
			land_bin[b] = np.sum(land[linx:i]*pdf_dis[linx:i])/total_land
			k_bin[b] = np.sum(ks[linx:i]*pdf_dis[linx:i])
			kl_bar[b] = k_bin[b]/land_bin[b]
			linx = i
			b +=1
	# Add last bin
	land_bin[b] = np.sum(land[linx:i]*y_prob[linx:i])/total_land
	k_bin[b] = np.sum(ks[linx:i]*y_prob[linx:i])
	kl_bar[b] = k_bin[b]/land_bin[b]

	kl_bar = (kl_bar/kl_bar[-1])[1:]

	return kl_bar

# PART III: Plotting modules # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # 

def plot_histogram(sol_R1,sol_P1,dataR,dataP,save_in=None):
	"""
	Produces a histogram comparing the results of the model with A&R (2014).
	It does so for the US calibration and the Poor country (Aggregate Factors) calibration.

	Inputs
	------
	sol_R1 (dic)	: a dictionary solution from solve(), rich country ("R") parameters.
	sol_P1 (dic)	: a dictionary solution from solve(), rich country ("R") parameters.
	dataR (np.array): array with the results from the original model: (y,pdf(y),l,k,Y) from calibration.m
	dataP (np.array): array with the results from the original model: (y,pdf(y),l,k,Y) from AGGFAC.m
	save_in (str)	: (optional) name of the file to save the plot.

	Output
	------
	A matplotlib plot.

	"""
	y_grid, y_prob, farm_grid, k_grid, output_m = dataR
	y_grid_p, y_prob_p, farm_grid_p, k_grid_p, output_m_p = dataP
	xs,thetas, ks, pdf = sol_R1['xs'],1/sol_R1['thetas'],sol_R1['ks'],sol_R1['pdf_dis']
	xs2,thetas2, ks2, pdf2 = sol_P1['xs'],1/sol_P1['thetas'],sol_P1['ks'],sol_P1['pdf_dis']

	bins = (0,4,20,28,40,57,73,89,105,202,405,809,4000)
	uniform = np.ones(6000)*1/6000
	pdf_dis = sol_R1['pdf_dis']
	plt.figure(figsize=(15,6))

	plt.subplot(1,2,1)
	plt.title("US Baseline", fontsize=16)
	h,b = np.histogram(thetas, bins, weights=pdf)
	hm,bm = np.histogram(farm_grid, bins, weights=y_prob)
	plt.bar(np.arange(len(h)),h,width=0.4,color=sn.xkcd_rgb["windows blue"],label="Matching")
	plt.bar(np.arange(len(hm))+0.5,hm,width=0.4,color=sn.xkcd_rgb["pale yellow"],label="A&R")

	plt.xticks(np.linspace(0.5,12.5,13), ('<4','4-20','20-28','28-40','40-57','57-73','73-89','89-105',
	           '105-202','202-405','405-809','+809'),rotation='vertical', fontsize=14)
	plt.yticks(fontsize=14)
	plt.xlim(0,12)
	plt.ylim(0,0.35)
	plt.legend(fontsize=16,frameon=True)

	plt.subplot(1,2,2)
	plt.title("Poor country (less TFP and Land)", fontsize=16)
	h2,b2 = np.histogram(thetas2, bins, weights=pdf2)
	hm2,bm2 = np.histogram(farm_grid_p, bins, weights=y_prob_p)
	plt.bar(np.arange(len(h2)),h2,width=0.4,color=sn.xkcd_rgb["windows blue"],label="Matching")
	plt.bar(np.arange(len(hm2))+0.5,hm2,width=0.4,color=sn.xkcd_rgb["pale yellow"],label="A&R")
	plt.xticks(np.linspace(0.5,12.5,13), ('<4','4-20','20-28','28-40','40-57','57-73','73-89','89-105',
	           '105-202','202-405','405-809','+809'),rotation='vertical', fontsize=14)
	plt.yticks(fontsize=16)
	plt.legend(fontsize=16,frameon=True)
	plt.xlim(0,12)

	plt.tight_layout()

	if save_in!=None:
		plt.savefig(save_in, dpi=None, facecolor='w', edgecolor='w',
		        orientation='portrait', papertype=None, format='pdf',
		        transparent=False, bbox_inches="tight", pad_inches=0.1,
		        frameon=None)
	plt.show()

def plot_land_choice(solsR,labelsR,solsP,labelsP,color=True,logs=True,save_in=None):
	"""
	Generates Report4.pdf and Report4b.pdf

	If soldP=None and logs=False, it delivers two plots: one with the normal scaling and one the log scaling.
	(Report4b.pdf)
	"""
	if color:
		sn.set_palette("colorblind",len(solsR)+2)
		alplist = np.ones(len(solsR))
	else:
		sn.set_palette("Greys_r",len(solsR)+2)
		alplist = np.ones(len(solsR))
	
	plt.figure(figsize=(12,6))
	allsols = zip(solsR,labelsR,alplist)
	plt.subplot(121)
	maxs = []
	mins = []
	for sol in allsols:
		plt.plot(sol[0]['xs'],1/sol[0]['thetas'],label=sol[1], alpha=sol[2])
		maxs.append(max(1/sol[0]['thetas']))
		mins.append(min(1/sol[0]['thetas'][sol[0]['xs']>=0.1]))
	plt.xlabel("farmer skill ($y$)", fontsize=16)
	plt.ylabel("farm size", fontsize=16)
	if logs:
		plt.xscale('log')
		plt.yscale('log')
		plt.xlim(0.1,100)
		plt.ylim(min(mins), max(maxs))
		tickies = np.logspace(np.log10(min(mins)), np.log10(max(maxs)),10)
		plt.yticks(tickies,np.round_(tickies, decimals=1),fontsize=12)
	else:
		plt.yticks(fontsize=12)
	# Same country plots
	if solsP==None:
		plt.subplot(122)
		maxs = []
		mins = []
		for sol in allsols:
			plt.plot(sol[0]['xs'],1/sol[0]['thetas'],label=sol[1], alpha=sol[2])
			maxs.append(max(1/sol[0]['thetas']))
			mins.append(min(1/sol[0]['thetas'][sol[0]['xs']>=0.1]))
		plt.xlabel("farmer skill ($y$)", fontsize=16)
		plt.ylabel("farm size", fontsize=16)
		plt.xscale('log')
		plt.yscale('log')
		plt.xlim(0.1,100)
		plt.ylim(min(mins), max(maxs))
		tickies = np.logspace(np.log10(min(mins)), np.log10(max(maxs)),10)
		plt.yticks(tickies,np.round_(tickies, decimals=1),fontsize=12)

	else: 
		allsols = zip(solsP,labelsP,alplist)
		plt.subplot(122)
		maxs = []
		mins = []
		for sol in allsols:
			plt.plot(sol[0]['xs'],1/sol[0]['thetas'],label=sol[1], alpha=sol[2])
			maxs.append(max(1/sol[0]['thetas']))
			mins.append(min(1/sol[0]['thetas'][sol[0]['xs']>=0.1]))
		if logs:
			plt.xscale('log')
			plt.yscale('log')
			plt.xlim(0.1,100)
			plt.ylim(min(mins), max(maxs))
			tickies = np.logspace(np.log10(min(mins)), np.log10(max(maxs)),10)
			plt.yticks(tickies,np.round_(tickies, decimals=1),fontsize=12)
		else:
			plt.yticks(fontsize=12)
	plt.xlabel("farmer skill ($y$)", fontsize=16)
	plt.ylabel("farm size", fontsize=16)
	plt.legend(bbox_to_anchor=(1.05, 1), loc=2, borderaxespad=0., frameon=True, fontsize=14)

	plt.tight_layout()
	if save_in != None:
		plt.savefig(save_in, dpi=None, facecolor='w', edgecolor='w',
		        orientation='portrait', papertype=None, format='pdf',
		        transparent=False, bbox_inches="tight", pad_inches=0.1,
		        frameon=None)

	plt.show()


# Other plotting modules (director's cut) # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# These plots are not in the paper but may be used in the notebook # # # # # # # # # # # # # # # # # # # # # # # 

def plot_kl_ratio(solsR,labelsR,solsP,labelsP,color=True,save_in=None, showw=True):
	"""
	Generates Report5.pdf
	"""
	if color:
		sn.set_palette("colorblind",len(solsR))
		alplist = np.ones(len(solsR))
	else:
		sn.set_palette("Greys_r",len(solsR)+2)
		alplist = np.ones(len(solsR))

	allsols = zip(solsR,labelsR,alplist)

	plt.figure(figsize=(12,6))
	plt.subplot(121)
	for sol in allsols:
		ratio = (sol[0]['ks']*sol[0]['thetas'])/((sol[0]['ks']*sol[0]['thetas'])[-1])
		plt.plot(sol[0]['xs'],ratio,label=sol[1], alpha=sol[2])
	plt.xlabel("Farmer Skill", fontsize=16)
	plt.ylabel('$k/l$', fontsize=16)
	plt.xticks(fontsize=14)
	plt.yticks(fontsize=14)
	plt.ylabel('$k/l$ normalized', fontsize=16)
	plt.xscale('log')

	allsols = zip(solsP,labelsP,alplist)
	plt.subplot(122)
	for sol in allsols:
		ratio = (sol[0]['ks']*sol[0]['thetas'])/((sol[0]['ks']*sol[0]['thetas'])[-1])
		plt.plot(sol[0]['xs'],ratio,label=sol[1], alpha=sol[2])
	plt.xlabel("Farmer Skill", fontsize=16)
	plt.ylabel('$k/l$', fontsize=16)
	plt.xticks(fontsize=14)
	plt.yticks(fontsize=14)
	plt.ylim(0,350)
	plt.ylabel('$k/l$ normalized', fontsize=16)
	plt.xscale('log')
	plt.legend(bbox_to_anchor=(1.05, 1), loc=2, borderaxespad=0., frameon=True, fontsize=14)

	if save_in != None:
		plt.savefig(save_in, dpi=None, facecolor='w', edgecolor='w',
		        orientation='portrait', papertype=None, format='pdf',
		        transparent=False, bbox_inches="tight", pad_inches=0.1,
		        frameon=None)
	if showw:
		plt.show()

def plot_dischange(sol_R1,sol_Rx,xrange=(0.9,1.1),extras=None,save_in=None,allplots=True,color=True,uc=cr, show=True):

	"""
	Plots the change in the distriution of firm size from an original distribution (sol_R1) to one with increased spread of x (sol_Rx).
	"""
	xs1,thetas1, ks1, pdf1 = sol_R1['xs'],1/sol_R1['thetas'],sol_R1['ks'],sol_R1['pdf_dis']
	xs10,thetas10, ks10, pdf10 = sol_Rx['xs'],1/sol_Rx['thetas'],sol_Rx['ks'],sol_Rx['pdf_dis']
	if extras != None:
		xs20,thetas20, ks20, pdf20 = extras[0]['xs'],1/extras[0]['thetas'],extras[0]['ks'],extras[0]['pdf_dis']

	if color==True:
		sty = ['-','-','-']
		c = uc
	else:
		sty = ['-','-','-']
		c = 'black'

	cdf10 = np.cumsum(pdf10)
	cdf1 = np.cumsum(pdf1)
	if extras != None:
		cdf20 = np.cumsum(pdf20)
	
	plt.figure(figsize=(12,6))
	
	plt.subplot(1,2,1)
	plt.plot(thetas1,pdf1,color=c, ls=sty[0],alpha=0.35)
	if extras != None:
		plt.plot(thetas20,pdf20,color=c, ls=sty[1],alpha=0.7)
	plt.plot(thetas10,pdf10,ls=sty[2],color=c)
	plt.ylabel('pdf', fontsize=18)
	plt.xlabel('farm size', fontsize=18)
	plt.xticks(fontsize=14)
	plt.yticks(fontsize=14)
	plt.xscale('log')

	plt.subplot(1,2,2)
	plt.plot(thetas1,cdf1,label='$x=1$',c=c,ls=sty[0],alpha=0.35)
	if extras != None:
		plt.plot(thetas20,cdf20,label='$x='+str(extras[1])+'$',c=c,ls=sty[1],alpha=0.7)
	plt.plot(thetas10,cdf10,label='$x='+str(xrange)+'$',ls=sty[2],c=c)
	plt.ylabel('cdf', fontsize=18)
	plt.xlabel('farm size', fontsize=18)
	plt.legend(bbox_to_anchor=(1.05, 1), loc=2, borderaxespad=0., frameon=True,fontsize=14)
	plt.xscale('log')
	plt.xticks(fontsize=14)
	plt.yticks(fontsize=14)
	plt.ylim(0,1)

	if save_in != None:
		plt.savefig(save_in, dpi=None, facecolor='w', edgecolor='w', orientation='portrait', papertype=None, 
			format='pdf', transparent=False, bbox_inches='tight', pad_inches=0.1, frameon=None)
	if show:
		plt.show()

def plot_k_fit(Grid,funcs,params,best_fit):
	"""
	Plots the fit of the approxiamted solution to k(x,y,l,r).

	Module used by get_k().

	"""
	# Unpacking
	Kgrid, xgrid, ygrid, lgrid = Grid
	params_ten3o,params_ten3,params_s = params
	res_ten3o, res_ten3, pol_shorcut = funcs
	cls = ['black','black','black']
	cls[best_fit] = 'green'

	#Plotting
	plt.figure(figsize=(12,3))
	plt.subplot(1,3,1)
	plt.plot(Kgrid.flatten(), color="grey", alpha=0.8)
	plt.plot(res_ten3o(params_ten3o,np.array([0.,]),xgrid,ygrid,lgrid), alpha=0.5)
	plt.title("Tensors 2nd",color=cls[0])
	plt.xlim(0,len(Kgrid.flatten()))

	plt.subplot(1,3,2)
	plt.plot(Kgrid.flatten(), color="grey", alpha=0.8)
	plt.plot(res_ten3(params_ten3,np.array([0.,]),xgrid,ygrid,lgrid), alpha=0.5)
	plt.title("Tensors 3rd",color=cls[1])
	plt.xlim(0,len(Kgrid.flatten()))

	plt.subplot(1,3,3)
	plt.plot(Kgrid.flatten(), color="grey", alpha=0.8)
	plt.plot(pol_shorcut(params_s,np.array([0.,]),xgrid.flatten(),ygrid.flatten(),lgrid.flatten()), alpha=0.5)
	plt.title("Polynomial approx",color=cls[2])
	plt.xlim(0,len(Kgrid.flatten()))

	plt.show()
