"""
Stand-alone version of the code to generate the plots for Assortative Matching with Large Firms (2016)
 
@author : Cristina Lafuente
@date : 31/05/2016

"""
#####################################################################
# Importing modules ################################################
# Classic libraries
from __future__ import division
import numpy as np
import sympy as sym
# Plotting libraries
import matplotlib.pyplot as plt
import seaborn as sn # library for pretty plots. See https://stanford.edu/~mwaskom/software/seaborn/ 
sn.set_style('whitegrid')
cp = sn.xkcd_rgb["pinkish"]
cb = "#3498db"
cr = "#1fa774"
# Specialist functions
from scipy.optimize import minimize, root, fsolve
from scipy.optimize import leastsq, least_squares
from scipy.interpolate import interp1d
from scipy.special import erf
# Core functions - solve the problem and plot the results
import capital

# Loading the data from A&R (2014)
CalUS = np.loadtxt('model_data.csv', delimiter=',').T
CalP = np.loadtxt('model_data_p.csv', delimiter=',').T
y_grid, y_prob, farm_grid, k_grid, output_m = CalUS
y_grid_p, y_prob_p, farm_grid_p, k_grid_p, output_m_p = CalP

# Solving and storing for k ##########################################
k_starsP25 = {}
k_starsP10 = {}
k_stars25 = {}
k_stars10 = {}
k_stars25[2] = capital.get_k("R", ftype=2,fchoice=1,plot=False, save_in="Rkpams25_{}".format(2), x_range=[0.75,1.25])
k_stars10[2] = capital.get_k("R", ftype=2,fchoice=0,plot=False, save_in="Rkpams{}".format(2), x_range=[0.9,1.1])
k_starsP25[2] = capital.get_k("P", ftype=2,fchoice=1,plot=False, save_in="Pkpams25_{}".format(2), x_range=[0.75,1.25])
k_starsP10[2] = capital.get_k("P", ftype=2,fchoice=0,plot=False, save_in="Pkpams{}".format(2), x_range=[0.9,1.1])
print 'Done with k'

# Solving original model (x=1) #######################################
sol_R1 = capital.solve("R", k_stars10[2], ftype=2,assort='positive',scaling_x=0.90607341901223637)
sol_P1 = capital.solve("P", k_starsP10[2], ftype=2,assort='positive',scaling_x=0.90607341901223637)
print 'Done with x=1'

# Increasing the Spread (lognormal only) ##############################

# Lognormal Distribution for x
_modules = [{'ImmutableMatrix': np.array, 'erf': erf, 'sqrt': np.sqrt}, 'numpy']
x, loc2, mu2, sigma2 = sym.var('x, loc2, mu2, sigma2')
productivity_cdf = (0.5 + 0.5 * sym.erf((sym.log(x - loc2) - mu2) / sym.sqrt(2 * sigma2**2)))
productivity_params62 = {'loc2': 0.0, 'mu2': 0.00, 'sigma2': 0.20}
# Symbolic version
pdf62 = productivity_cdf.diff(x).subs(productivity_params62)
# Executable version
pdf_exe62 = sym.lambdify(x, pdf62, 'numpy')
cdf_exe62 = sym.lambdify(x, productivity_cdf.subs(productivity_params62), _modules)
# Normalization constants (so total land = 1)
norm62 = 1/(cdf_exe62(1.1)-cdf_exe62(0.9))
c62= 1.430280192432233*norm62-1

## Solving
sol_log = {}
sol_logP = {}
# Spread [0.95,1.05]
sol_logP[2605]= capital.solve("P", k_starsP10[2], ftype=2,assort='positive',spread='lognorm',dispams=[0.00,0.2],
                           x_range=[0.95,1.05],scaling_x=norm62*1.7611607186330467, normconst=c62)
# Spread [0.90,1.10]
sol_logP[2610]= capital.solve("P", k_starsP10[2], ftype=2,assort='positive',spread='lognorm',dispams=[0.00,0.2],
                             x_range=[0.9,1.1],scaling_x=norm62*0.9061110680375417, normconst=c62)
# Spread [0.85,1.15]
sol_logP[2615]= capital.solve("P", k_starsP25[2], ftype=2,assort='positive',spread='lognorm',dispams=[0.00,0.2],
                             x_range=[0.85,1.15],scaling_x=1.6494247873450758, normconst=c62)
# Spread [0.80,1.20]
sol_logP[2620]= capital.solve("P", k_starsP25[2], ftype=2,assort='positive',spread='lognorm',dispams=[0.00,0.2],
                             x_range=[0.8,1.20],scaling_x=1.3196670116025528, normconst=c62)
# Spread [0.75,1.25]
sol_logP[2625]= capital.solve("P", k_starsP25[2], ftype=2,assort='positive',spread='lognorm',dispams=[0.00,0.2],
                             x_range=[0.75,1.25],scaling_x=1.1434977494042775, normconst=c62)

# PLOTING ##################################################################

# Histogram - Comparative with A&R (2014)
capital.plot_histogram(sol_R1,sol_P1,CalUS,CalP,save_in='Figure7.pdf')

# Land choice plot
solsP = (sol_P1,sol_logP[2605],sol_logP[2610],sol_logP[2615],sol_logP[2620],sol_logP[2625])
labs = ('$x=0$','$x\in[0.95,1.05]$ ','$x\in[0.90,1.10]$','$x\in[0.85,1.15]$','$x\in[0.80,1.20]$','$x\in[0.75,1.25]$')

capital.plot_land_choice(solsP,labs,None,None, color=True, logs=False,save_in='Figure1.pdf')

