# Classic libraries
from __future__ import division
import numpy as np
import sympy as sym
# Plotting libraries
import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D
import seaborn as sn # library for pretty plots. See https://stanford.edu/~mwaskom/software/seaborn/ 
sn.set_style('whitegrid')
cp = sn.xkcd_rgb["pinkish"]
cb = "#3498db"
cr = "#1fa774"
from matplotlib import cm
from matplotlib.ticker import LinearLocator, FormatStrFormatter
import matplotlib.colors as colors

# Specialist functions
from scipy.optimize import minimize, root, fsolve
from scipy.optimize import leastsq #, least_squares
from scipy.interpolate import interp1d
from scipy.special import erf

# Specific modules
import inputs
import models
import shooting


# define symbolic expression for CES between x and y
x, y, omega_A, sigma_A = sym.var('x, y, omega_A, sigma_A')
A = ((omega_A * x**((sigma_A - 1) / sigma_A) + 
     (1 - omega_A) * y**((sigma_A - 1) / sigma_A))**(sigma_A / (sigma_A - 1))) 

# define symbolic expression for CES between x and y
r, l, omega_B, sigma_B = sym.var('r, l, omega_B, sigma_B')
B = l**omega_B * r**(1-omega_B)

F = A * B

# positive assortativity requires that sigma_A * sigma_B < 1
F_params = {'omega_A':0.25, 'omega_B':0.5, 'sigma_A':0.5, 'sigma_B':1.0 }
F_exe = sym.lambdify((x,y,l,r),F.subs(F_params))
x_bounds = [1.0, 1e2]
y_bounds = [1.0, 1e2]

# Parameters and distibutions
xun_params = {'aw': x_bounds[0], 'bw': x_bounds[1]}
yun_params = {'af': y_bounds[0], 'bf': y_bounds[1]}
aw, af, bw, bf = sym.var('aw, af, bw, bf')
cdf_wuni = (x - aw) / (bw - aw)
cdf_funi = (y - af) / (bf - af)
wuni = inputs.Input(var=x,cdf=cdf_wuni*100,params=xun_params,bounds=x_bounds)
funi = inputs.Input(var=y,cdf=cdf_funi,params=yun_params,bounds=y_bounds)

# Number of knots in the grid
nk =6000
kts = None

model = models.Model(assortativity='positive',
                     workers=wuni,firms=funi,
                     production=F,params=F_params)

# PAM # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

# Omega A # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

# Space to store coefficients
sol_omA = np.empty((9,6000,5))
# Range for omega_a
oms = np.arange(0.1,1.0,0.1)

# Solve
for i, omA in enumerate(oms):
    F_params = {'omega_A':omA, 'omega_B':0.5, 'sigma_A':0.5, 'sigma_B':1.0 }
    model = models.Model(assortativity='positive',workers=wuni,firms=funi,production=F,
                         params=F_params)
    solver = shooting.ShootingSolver(model=model)
    # Solve
    solver.solve(2000.0, tol=1e-6, number_knots=nk, knots=kts,integrator="lsoda",message=False)
    sol = solver.solution
    sol_i = np.empty((6000,5))
    sol_i[:,0] = sol.index.values
    sol_i[:,1] = sol['$\\mu(x)$'].values
    sol_i[:,2] = sol['$\\theta(x)$'].values
    sol_i[:,3] = sol['$w(x)$'].values
    sol_i[:,4] = sol['$\\pi(x)$'].values
    sol_omA[i] = sol_i

# Plot

plt.figure(figsize=(15,5))

# select colorscheme (uncomment for color)
sn.set_palette("Greys_r",10)
#sn.set_palette('rainbow',13)

# Line thickness and style
lws = [4,3,3,2,2]
lend = ['butt','round','round','round','round']
sts = [':',':','-.','--','-']
# Uncomment below for solid lines
#sts = ['-','-','-','-','-']

for i, sol in enumerate(sol_omA[0:10:2]):
    xs, ys, thetas = sol[:,0],sol[:,1],sol[:,2]
    plt.subplot(131)
    plt.plot(xs, ys, label=oms[i*2],ls=sts[i],dash_capstyle=lend[i],lw=lws[i])
    plt.xlabel(r'$x$', fontsize=20)
    plt.ylabel(r'$\mu(x)$', fontsize=20)
    
    plt.subplot(132)
    plt.plot(xs, thetas, label=oms[i*2],ls=sts[i],dash_capstyle=lend[i],lw=lws[i])
    plt.ylim(0,1000)
    plt.xlabel(r'$x$', fontsize=20)
    plt.ylabel(r'$\theta(x)$', fontsize=20)
    plt.legend(loc='best',frameon=True, title=r'$\omega_A$')
    
    plt.subplot(133)
    plt.plot(ys, thetas, label=oms[i*2],ls=sts[i],dash_capstyle=lend[i],lw=lws[i])
    plt.ylim(0,1000)
    plt.xlabel(r'$\mu(x)$', fontsize=20)
    plt.ylabel(r'$\theta(\mu(x))$', fontsize=20)

plt.tight_layout()
plt.savefig('omA.pdf',format='pdf',dpi=1000,bbox_inches='tight')
# uncomment for show
#plt.show()

# Omega B # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

sol_omB = np.empty((9,6000,5))
oms = np.arange(0.1,1.,0.1)
guess = [240,550,550,168,168,168,168,200,400] # Solver unstable, use guesses provided
for i, omB in enumerate(oms):
    F_params = {'omega_A':0.5, 'omega_B':omB, 'sigma_A':0.5, 'sigma_B':1.0 }
    model = models.Model(assortativity='positive',workers=wuni,firms=funi,production=F,
                         params=F_params)
    solver = shooting.ShootingSolver(model=model)
    # Solve
    solver.solve(guess[i], tol=1e-6, number_knots=nk, knots=kts,integrator="lsoda",message=False)
    sol = solver.solution
    sol_i = np.empty((6000,5))
    sol_i[:,0] = sol.index.values
    sol_i[:,1] = sol['$\\mu(x)$'].values
    sol_i[:,2] = sol['$\\theta(x)$'].values
    sol_i[:,3] = sol['$w(x)$'].values
    sol_i[:,4] = sol['$\\pi(x)$'].values
    sol_omB[i] = sol_i

# Plot

plt.figure(figsize=(15,5))
for i, sol in enumerate(sol_omB[0:10:2]):
    xs, ys, thetas = sol[:,0],sol[:,1],sol[:,2]
    plt.subplot(131)
    plt.plot(xs, ys, label=oms[i*2],ls=sts[i],dash_capstyle=lend[i],lw=lws[i])
    plt.xlabel(r'$x$', fontsize=20)
    plt.ylabel(r'$\mu(x)$', fontsize=20)
    
    plt.subplot(132)
    plt.plot(xs, thetas, label=oms[i*2],ls=sts[i],dash_capstyle=lend[i],lw=lws[i]) #c= scalarMap.to_rgba(cvals[i])
    plt.ylim(0,1000)
    plt.xlabel(r'$x$', fontsize=20)
    plt.ylabel(r'$\theta(x)$', fontsize=20)
    plt.legend(loc='best',frameon=True, title=r'$\omega_{ \theta}$')
    
    plt.subplot(133)
    plt.plot(ys, thetas, label=oms[i*2],ls=sts[i],dash_capstyle=lend[i],lw=lws[i])
    plt.ylim(0,1000)
    plt.xlabel(r'$\mu(x)$', fontsize=20)
    plt.ylabel(r'$\theta(\mu(x))$', fontsize=20)

plt.tight_layout()
plt.savefig('omB.pdf',format='pdf',dpi=1000,bbox_inches='tight',)
plt.show()


# Sigma_a  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

sol_sigA = np.empty((9,6000,5))
sigA_positive = np.arange(0.2,1.2,0.2)
for i,sigA in enumerate(sigA_positive):
    F_params = {'omega_A':0.6, 'omega_B':0.3, 'sigma_A':sigA, 'sigma_B':1.0 }
    model = models.Model(assortativity='positive',workers=wuni,firms=funi,production=F,
                             params=F_params)
    solver = shooting.ShootingSolver(model=model)
    solver.solve(50000., tol=1e-6, number_knots=nk, knots=kts,integrator='dopri5',message=False)
    sol = solver.solution
    sol_i = np.empty((6000,5))
    sol_i[:,0] = sol.index.values
    sol_i[:,1] = sol['$\\mu(x)$'].values
    sol_i[:,2] = sol['$\\theta(x)$'].values
    sol_i[:,3] = sol['$w(x)$'].values
    sol_i[:,4] = sol['$\\pi(x)$'].values
    sol_sigA[i] = sol_i

# Plot

plt.figure(figsize=(15,5))
for i, sol in enumerate(sol_sigA[:5]):
    plt.subplot(131)
    xs, ys, thetas = sol[:,0],sol[:,1],sol[:,2]
    plt.plot(xs, ys, label=sigA_positive[i],ls=sts[i],dash_capstyle=lend[i],lw=lws[i])
    plt.xlim(0,100)
    plt.ylabel(r'$\mu(x)$', fontsize=20)
    plt.xlabel(r'$x$', fontsize=20)    
    
    plt.subplot(132)
    plt.plot(xs, thetas, label=sigA_positive[i],ls=sts[i],dash_capstyle=lend[i],lw=lws[i])
    plt.ylim(0,500)
    plt.xlim(0,100)
    plt.xlabel(r'$x$', fontsize=20)
    plt.ylabel(r'$\theta(x)$', fontsize=20)
    plt.legend(loc='best',frameon=True,title="$\sigma_A$")
    
    plt.subplot(133)
    plt.plot(ys, thetas, label=sigA_positive[i],ls=sts[i],dash_capstyle=lend[i],lw=lws[i])
    plt.ylim(0,500)
    plt.xlabel(r'$\mu(x)$', fontsize=20)
    plt.ylabel(r'$\theta(\mu(x))$', fontsize=20)

plt.tight_layout()
plt.savefig('sigA.pdf',format='pdf',dpi=1000,bbox_inches='tight',)

# NAM # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

# Omega A # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

nam_sols_omA = np.zeros((5,6000,3))
omArange = [0.1,0.3,0.5,0.7,0.9]

# Here the initial guesses needs to change for the solver to end successfully
guesses = [1e4,1e4,1e5,1e5,1e6]

for i, omA in enumerate(omArange):
    F_params = {'omega_A':omA, 'omega_B':0.5, 'sigma_A':1.1, 'sigma_B':1.0 }
    model = models.Model(assortativity='negative',workers=wuni,firms=funi,production=F,params=F_params)
    solver = shooting.ShootingSolver(model=model)
    solver.solve(guesses[i], tol=1e-6, number_knots=nk, knots=kts,integrator='dopri5',message=False)
    sol = solver.solution
    nam_sols_omA[i,:,0] = sol.index.values
    nam_sols_omA[i,:,1] = sol['$\\mu(x)$'].values
    nam_sols_omA[i,:,2] = sol['$\\theta(x)$'].values

# Plot

plt.figure(figsize=(15,5))
for i,sol in enumerate(nam_sols_omA ):
    xs, ys, ths = sol[:,0],sol[:,1],sol[:,2]
    plt.subplot(131)
    plt.plot(xs,ys,label=omArange[i],ls=sts[i],dash_capstyle=lend[i],lw=lws[i])
    plt.xlabel('$x$', fontsize=20)
    plt.ylabel(r'$\mu(x)$', fontsize=20)
    plt.xlim(0,100)   

    plt.subplot(132)
    plt.plot(ys,ths,label=omArange[i],ls=sts[i],dash_capstyle=lend[i],lw=lws[i])
    plt.xlabel(r'$x$', fontsize=20)
    plt.ylabel(r'$\theta(x)$', fontsize=20)
    plt.legend(loc='best',frameon=True,title=r'$\omega_A$',fontsize=12)
    plt.xlim(0,100)
    plt.ylim(0,1000)

    plt.subplot(133)
    plt.plot(xs,ths,label=omArange[i],ls=sts[i],dash_capstyle=lend[i],lw=lws[i])
    plt.xlabel(r'$\mu(x)$', fontsize=20)
    plt.ylabel(r'$\theta(\mu(x))$', fontsize=20)
    plt.xlim(0,100)
    plt.ylim(0,1000)

plt.tight_layout()
plt.savefig('omA_NAM.pdf',format='pdf',dpi=1000,bbox_inches='tight',)

# Omega B # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

nam_sols_omB = np.zeros((5,6000,3))

# Here the parameter range for omega_B needs to be split into two.
# This insures the shooting solver finishes successfully.
omBrange = [0.1,0.3,0.5]
for i,omB in enumerate(omBrange):
    F_params = {'omega_A':0.5, 'omega_B':omB, 'sigma_A':1.1, 'sigma_B':1.0 }
    model = models.Model(assortativity='negative',workers=wuni,firms=funi,production=F,params=F_params)
    solver = shooting.ShootingSolver(model=model)
    solver.solve(5e10, tol=1e-6, number_knots=nk, knots=kts,integrator='lsoda',message=False)
    sol = solver.solution
    nam_sols_omB[i,:,0] = sol.index.values
    nam_sols_omB[i,:,1] = sol['$\\mu(x)$'].values
    nam_sols_omB[i,:,2] = sol['$\\theta(x)$'].values

omBrange = [0.7,0.9]
for i,omB in enumerate(omBrange):
    F_params = {'omega_A':0.5, 'omega_B':0.9, 'sigma_A':1.1, 'sigma_B':1.0 }
    model = models.Model(assortativity='negative',workers=wuni,firms=funi,production=F,params=F_params)
    solver = shooting.ShootingSolver(model=model)
    solver.solve(5999., tol=1e-4, number_knots=nk, knots=kts,integrator='lsoda',message=False)
    sol = solver.solution
    nam_sols_omB[i+3,:,0] = sol.index.values
    nam_sols_omB[i+3,:,1] = sol['$\\mu(x)$'].values
    nam_sols_omB[i+3,:,2] = sol['$\\theta(x)$'].values

# Plot

omBrange = [0.1,0.3,0.5,0.7,0.9]
plt.figure(figsize=(15,5))
for i,sol in enumerate(nam_sols_omB[:-1]):
    xs, ys, ths = sol[:,0],sol[:,1],sol[:,2]
    plt.subplot(131)
    plt.plot(xs,ys,label=omBrange[i],ls=sts[i],dash_capstyle=lend[i],lw=lws[i])
    plt.xlabel('$x$', fontsize=20)
    plt.ylabel(r'$\mu(x)$', fontsize=20)
    plt.xlim(0,100)   

    plt.subplot(132)
    plt.plot(ys,ths,label=omBrange[i],ls=sts[i],dash_capstyle=lend[i],lw=lws[i])
    plt.xlabel(r'$x$', fontsize=20)
    plt.ylabel(r'$\theta(x)$', fontsize=20)
    plt.legend(loc='best',frameon=True,title=r'$\omega_{\theta}$',fontsize=12)
    plt.xlim(0,100)
    plt.ylim(0,1000)

    plt.subplot(133)
    plt.plot(xs,ths,label=omBrange[i],ls=sts[i],dash_capstyle=lend[i],lw=lws[i])
    plt.xlabel(r'$\mu(x)$', fontsize=20)
    plt.ylabel(r'$\theta(\mu(x))$', fontsize=20)
    plt.xlim(0,100)
    plt.ylim(0,1000)

plt.tight_layout()
plt.savefig('omB_NAM.pdf',format='pdf',dpi=1000,bbox_inches='tight',)

# Sigma_a  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

sigArange = [1.1,3.0,5.0,9.0,50.0]
nam_sols_sigA = np.zeros((5,6000,3))
for i, sigA in enumerate(sigArange):
    F_params = {'omega_A':0.5, 'omega_B':0.5, 'sigma_A':sigA, 'sigma_B':1.0 }
    model = models.Model(assortativity='negative',workers=wuni,firms=funi,production=F,params=F_params)
    solver = shooting.ShootingSolver(model=model)
    solver.solve(1e5, tol=1e-6, number_knots=nk, knots=kts,integrator='lsoda',message=False)
    sol = solver.solution
    nam_sols_sigA[i,:,0] = sol.index.values
    nam_sols_sigA[i,:,1] = sol['$\\mu(x)$'].values
    nam_sols_sigA[i,:,2] = sol['$\\theta(x)$'].values

# Plot

plt.figure(figsize=(15,5))
for i,sol in enumerate(nam_sols_sigA ):
    xs, ys, ths = sol[:,0],sol[:,1],sol[:,2]
    plt.subplot(131)
    plt.plot(xs,ys,label=sigArange[i],ls=sts[i],dash_capstyle=lend[i],lw=lws[i])
    plt.xlabel('$x$', fontsize=20)
    plt.ylabel(r'$\mu(x)$', fontsize=20)
    plt.xlim(0,100)   

    plt.subplot(132)
    plt.plot(ys,ths,label=sigArange[i],ls=sts[i],dash_capstyle=lend[i],lw=lws[i])
    plt.xlabel(r'$x$', fontsize=20)
    plt.ylabel(r'$\theta(x)$', fontsize=20)
    plt.legend(loc='best',frameon=True,title=r'$\sigma_A$',fontsize=12)
    plt.xlim(0,100)
    plt.ylim(0,1000)

    plt.subplot(133)
    plt.plot(xs,ths,label=sigArange[i],ls=sts[i],dash_capstyle=lend[i],lw=lws[i])
    plt.xlabel(r'$\mu(x)$', fontsize=20)
    plt.ylabel(r'$\theta(\mu(x))$', fontsize=20)
    plt.xlim(0,100)
    plt.ylim(0,1000)

plt.tight_layout()
plt.savefig('sigA_NAM.pdf',format='pdf',dpi=1000,bbox_inches='tight',)

